/* client_keyexchange.c
 *
 * $Id: client_keyexchange.c,v 1.47 2000/06/01 17:01:54 nisse Exp $ */

/* lsh, an implementation of the ssh protocol
 *
 * Copyright (C) 1998 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "client_keyexchange.h"

#include "atoms.h"
#include "command.h"
#include "debug.h"
#include "format.h"
#include "lookup_verifier.h"
#include "password.h"
#include "srp.h"
#include "ssh.h"
#include "werror.h"
#include "xalloc.h"

#include <assert.h>

#include "client_keyexchange.c.x"

/* GABA:
   (class
     (name dh_client_exchange)
     (super keyexchange_algorithm)
     (vars
       (dh object dh_method)))
*/

/* Handler for the kex_dh_reply message */
/* GABA:
   (class
     (name dh_client)
     (super packet_handler)
     (vars
       (dh struct dh_instance)
       (hostkey_algorithm . UINT32)
       (verifier object lookup_verifier)
       (algorithms object object_list)))
*/
    
static void
do_handle_dh_reply(struct packet_handler *c,
		   struct ssh_connection *connection,
		   struct lsh_string *packet)
{
  CAST(dh_client, closure, c);
  struct verifier *v;
  struct lsh_string *hash;
  int res;
  
  trace("handle_dh_reply()\n");

  res = dh_process_server_msg(&closure->dh, packet);
  lsh_string_free(packet);
  if (!res)
    {
      disconnect_kex_failed(connection, "Bad dh-reply\r\n");
      return;
    }
    
  v = LOOKUP_VERIFIER(closure->verifier, closure->hostkey_algorithm,
		      NULL, closure->dh.server_key);

  if (!v)
    {
      /* FIXME: Use a more appropriate error code? */
      disconnect_kex_failed(connection, "Bad server host key\r\n");
      return;
    }

#if DATAFELLOWS_WORKAROUNDS
  if (closure->hostkey_algorithm == ATOM_SSH_DSS && 
      (connection->peer_flags & PEER_SSH_DSS_KLUDGE))
    {
      v = make_dsa_verifier_kludge(v);
    }
#endif

  if (!dh_verify_server_msg(&closure->dh, v))
    {
      /* FIXME: Use a more appropriate error code? */
      disconnect_kex_failed(connection, "Invalid server signature\r\n");
      return;
    }
  
  /* Key exchange successful! */

  connection->dispatch[SSH_MSG_KEXDH_REPLY] = connection->fail;

  hash = closure->dh.exchange_hash;
  closure->dh.exchange_hash = NULL; /* For gc */

  keyexchange_finish(connection, closure->algorithms,
		     closure->dh.method->H,
		     hash,
		     ssh_format("%n", closure->dh.K));
}

static void
do_init_client_dh(struct keyexchange_algorithm *c,
		  struct ssh_connection *connection,
		  int hostkey_algorithm_atom,
		  struct lsh_object *extra,
		  struct object_list *algorithms)
{
  CAST(dh_client_exchange, closure, c);
  CAST_SUBTYPE(lookup_verifier, verifier, extra);
  
  NEW(dh_client, dh);

  assert(verifier);
  
  /* Initialize */
  dh->super.handler = do_handle_dh_reply;
  init_dh_instance(closure->dh, &dh->dh, connection);

  dh->verifier = verifier;
  dh->hostkey_algorithm = hostkey_algorithm_atom;

  dh->algorithms = algorithms;
  
  /* Send client's message */
  C_WRITE(connection, dh_make_client_msg(&dh->dh));

  /* Install handler */
  connection->dispatch[SSH_MSG_KEXDH_REPLY] = &dh->super;
  
  connection->kex_state = KEX_STATE_IN_PROGRESS;
}


struct keyexchange_algorithm *
make_dh_client(struct dh_method *dh)
{
  NEW(dh_client_exchange, self);

  self->super.init = do_init_client_dh;
  self->dh = dh;
  
  return &self->super;
}


#if WITH_SRP
/* SRP key exchange */

/* GABA:
   (class
     (name srp_client_instance)
     (vars
       (dh struct dh_instance)
       (name string)
       ;; (salt string)
       (algorithms object object_list)))
*/

/* GABA:
   (class
     (name srp_client_handler)
     (super packet_handler)
     (vars
       (srp object srp_client_instance)))
*/

static void
do_srp_client_proof_handler(struct packet_handler *s,
			    struct ssh_connection *connection,
			    struct lsh_string *packet)
{
  CAST(srp_client_handler, self, s);

  int res = srp_process_server_proof(&self->srp->dh, packet);
  lsh_string_free(packet);

  connection->dispatch[SSH_MSG_KEXSRP_PROOF] = connection->fail;
  
  if (res)
    {
      struct lsh_string *hash;

      hash = self->srp->dh.exchange_hash;
      self->srp->dh.exchange_hash = NULL; /* For gc */

      connection->flags |= CONNECTION_SRP;
      
      keyexchange_finish(connection, self->srp->algorithms,
			 self->srp->dh.method->H,
			 hash,
			 ssh_format("%n", self->srp->dh.K));
    }
  else
    PROTOCOL_ERROR(connection->e, "Invalid SSH_MSG_KEXSRP_PROOF message");

  /* Try to purge information about the key exchange */
  KILL(self->srp);
  KILL(self);
}

static struct packet_handler *
make_srp_client_proof_handler(struct srp_client_instance *srp)
{
  NEW(srp_client_handler, self);
  self->super.handler = do_srp_client_proof_handler;
  self->srp = srp;

  return &self->super;
}

#define MAX_PASSWD 100

static void
do_handle_srp_reply(struct packet_handler *s,
		    struct ssh_connection *connection,
		    struct lsh_string *packet)
{
  CAST(srp_client_handler, self, s);
  struct lsh_string *salt = srp_process_reply_msg(&self->srp->dh, packet);
  struct lsh_string *passwd;
  struct lsh_string *response;
  
  mpz_t x;
  
  connection->dispatch[SSH_MSG_KEXSRP_REPLY] = connection->fail;
    
  if (!salt)
    {
      PROTOCOL_ERROR(connection->e, "Invalid SSH_MSG_KEXSRP_REPLY message");
      return;
    }

  /* Ask for SRP password */
  passwd = read_password(MAX_PASSWD,
			 ssh_format("SRP password for %lS: ",
				    self->srp->name), 1);
  if (!passwd)
    {
      lsh_string_free(salt);
      disconnect_kex_failed(connection, "Bye");
    }
  
  mpz_init(x);
  srp_hash_password(x, self->srp->dh.method->H,
		    salt, self->srp->name,
		    passwd);

  lsh_string_free(salt);
  lsh_string_free(passwd);

  response = srp_make_client_proof(&self->srp->dh, x);
  mpz_clear(x);

  if (!response)
    PROTOCOL_ERROR(connection->e,
		   "SRP failure: Invalid public value from server.");
  
  C_WRITE(connection, response);
  
  connection->dispatch[SSH_MSG_KEXSRP_PROOF]
    = make_srp_client_proof_handler(self->srp);
}

static struct packet_handler *
make_srp_reply_handler(struct srp_client_instance *srp)
{
  NEW(srp_client_handler, self);
  self->super.handler = do_handle_srp_reply;
  self->srp = srp;

  return &self->super;
}

/* GABA:
   (class
     (name srp_client_exchange)
     (super keyexchange_algorithm)
     (vars
       (dh object dh_method)
       (name string)))
*/

static void
do_init_client_srp(struct keyexchange_algorithm *s,
		   struct ssh_connection *connection,
		   int hostkey_algorithm_atom UNUSED,
		   struct lsh_object *extra UNUSED,
		   struct object_list *algorithms)
{
  CAST(srp_client_exchange, self, s);
  
  NEW(srp_client_instance, srp);

  /* Initialize */
  init_dh_instance(self->dh, &srp->dh, connection);

  srp->algorithms = algorithms;
  srp->name = self->name;
  
  /* Send client's message */
  C_WRITE(connection, srp_make_init_msg(&srp->dh, self->name));

  /* Install handler */
  connection->dispatch[SSH_MSG_KEXSRP_REPLY] = make_srp_reply_handler(srp);
  
  connection->kex_state = KEX_STATE_IN_PROGRESS;
}

struct keyexchange_algorithm *
make_srp_client(struct dh_method *dh, struct lsh_string *name)
{
  NEW(srp_client_exchange, self);

  assert(dh->G->add);
  assert(dh->G->subtract);

  self->super.init = do_init_client_srp;
  self->dh = dh;
  self->name = name;
  
  return &self->super;
}

#endif /* WITH_SRP */
